#!/usr/bin/env python3

import torch.nn as nn
import torch as T
import torch.nn.functional as F

from torch.nn.modules.rnn import RNNCellBase
from subLSTM.functional import SubLSTMCell as SubLSTMCellF

import math


class SubLSTMCell(RNNCellBase):
  r"""A long sub-short-term memory (subLSTM) cell, as described in the paper:
  https://arxiv.org/abs/1711.02448

  .. math::

    \begin{array}{ll}
    i = \mathrm{sigmoid}(W_{ii} x + b_{ii} + W_{hi} h + b_{hi}) \\
    f = \mathrm{sigmoid}(W_{if} x + b_{if} + W_{hf} h + b_{hf}) \\
    g = \mathrm{sigmoid}(W_{ig} x + b_{ig} + W_{hc} h + b_{hg}) \\
    o = \mathrm{sigmoid}(W_{io} x + b_{io} + W_{ho} h + b_{ho}) \\
    c' = f * c + g - i \\
    h' = \mathrm{sigmoid}(c') - o \\
    \end{array}

  Args:
    input_size: The number of expected features in the input x
    hidden_size: The number of features in the hidden state h
    bias: If `False`, then the layer does not use bias weights `b_ih` and
      `b_hh`. Default: True

  Inputs: input, (h_0, c_0)
    - **input** (batch, input_size): tensor containing input features
    - **h_0** (batch, hidden_size): tensor containing the initial hidden
      state for each element in the batch.
    - **c_0** (batch. hidden_size): tensor containing the initial cell state
      for each element in the batch.

  Outputs: h_1, c_1
    - **h_1** (batch, hidden_size): tensor containing the next hidden state
      for each element in the batch
    - **c_1** (batch, hidden_size): tensor containing the next cell state
      for each element in the batch

  Attributes:
    weight_ih: the learnable input-hidden weights, of shape
      `(4*hidden_size x input_size)`
    weight_hh: the learnable hidden-hidden weights, of shape
      `(4*hidden_size x hidden_size)`
    bias_ih: the learnable input-hidden bias, of shape `(4*hidden_size)`
    bias_hh: the learnable hidden-hidden bias, of shape `(4*hidden_size)`

  Examples::

    >>> rnn = nn.SubLSTMCell(10, 20)
    >>> input = Variable(torch.randn(6, 3, 10))
    >>> hx = Variable(torch.randn(3, 20))
    >>> cx = Variable(torch.randn(3, 20))
    >>> output = []
    >>> for i in range(6):
    ...   hx, cx = rnn(input[i], (hx, cx))
    ...   output.append(hx)
  """

  def __init__(self, input_size, hidden_size, bias=True):
    super(SubLSTMCell, self).__init__()
    self.input_size = input_size
    self.hidden_size = hidden_size
    self.bias = bias
    self.weight_ih = nn.Parameter(T.Tensor(4 * hidden_size, input_size))
    self.weight_hh = nn.Parameter(T.Tensor(4 * hidden_size, hidden_size))
    if bias:
      self.bias_ih = nn.Parameter(T.Tensor(4 * hidden_size))
      self.bias_hh = nn.Parameter(T.Tensor(4 * hidden_size))
    else:
      self.register_parameter('bias_ih', None)
      self.register_parameter('bias_hh', None)
    self.reset_parameters()

  def reset_parameters(self):
    stdv = 1.0 / math.sqrt(self.hidden_size)
    for weight in self.parameters():
      weight.data.uniform_(-stdv, stdv)

  def forward(self, input, hx):
    return SubLSTMCellF(
        input, hx,
        self.weight_ih, self.weight_hh,
        self.bias_ih, self.bias_hh,
    )



#!/usr/bin/env python3

import torch.nn as nn
import torch as T
import torch.nn.functional as F


def SubLSTMCell(input, hidden, w_ih, w_hh, b_ih=None, b_hh=None):

  hx, cx = hidden
  gates = F.linear(input, w_ih, b_ih) + F.linear(hx, w_hh, b_hh)

  ingate, forgetgate, cellgate, outgate = gates.chunk(4, 1)

  ingate = F.sigmoid(ingate)
  forgetgate = F.sigmoid(forgetgate)
  cellgate = F.sigmoid(cellgate)
  outgate = F.sigmoid(outgate)

  cy = (forgetgate * cx) + (cellgate - ingate)
  hy = F.sigmoid(cy) - outgate

  return hy, cy



#!/usr/bin/env python3

import torch.nn as nn
import torch as T
import torch.nn.functional as F
from torch.autograd import Variable as var
from torch.nn import Module

from torch.nn.modules.rnn import RNNCellBase
from torch.nn.utils.rnn import PackedSequence, pack_padded_sequence as pack, pad_packed_sequence as pad

from subLSTM.functional import SubLSTMCell as SubLSTMCellF

import math


class SubLSTM(Module):

  def __init__(
      self,
      input_size,
      hidden_size,
      num_layers=1,
      bias=True,
      batch_first=False,
      dropout=0,
      bidirectional=False
  ):
    super(SubLSTM, self).__init__()
    self.input_size = input_size
    self.hidden_size = hidden_size
    self.num_layers = num_layers
    self.bias = bias
    self.batch_first = batch_first
    self.dropout = dropout
    self.dropout_state = {}
    self.bidirectional = bidirectional
    num_directions = 2 if bidirectional else 1

    gate_size = 4 * hidden_size

    self._all_weights = []
    for layer in range(num_layers):
      for direction in range(num_directions):
        layer_input_size = input_size if layer == 0 else hidden_size * num_directions

        w_ih = nn.Parameter(T.Tensor(gate_size, layer_input_size))
        w_hh = nn.Parameter(T.Tensor(gate_size, hidden_size))
        b_ih = nn.Parameter(T.Tensor(gate_size))
        b_hh = nn.Parameter(T.Tensor(gate_size))
        layer_params = (w_ih, w_hh, b_ih, b_hh)

        suffix = '_reverse' if direction == 1 else ''
        param_names = ['weight_ih_l{}{}', 'weight_hh_l{}{}']
        if bias:
          param_names += ['bias_ih_l{}{}', 'bias_hh_l{}{}']
        param_names = [x.format(layer, suffix) for x in param_names]

        for name, param in zip(param_names, layer_params):
          setattr(self, name, param)
        self._all_weights.append(param_names)

    self.flatten_parameters()
    self.reset_parameters()

  def flatten_parameters(self):
    pass

  def _apply(self, fn):
    ret = super(SubLSTM, self)._apply(fn)
    self.flatten_parameters()
    return ret

  def reset_parameters(self):
    stdv = 1.0 / math.sqrt(self.hidden_size)
    for weight in self.parameters():
      weight.data.uniform_(-stdv, stdv)

  def forward(self, input, hx=None):
    timesteps = input.size(1) if self.batch_first else input.size(0)
    directions = 2 if self.bidirectional else 1
    is_packed = isinstance(input, PackedSequence)

    if is_packed:
      input, batch_sizes = pad(input)
      max_batch_size = batch_sizes[0]
    else:
      batch_sizes = None
      max_batch_size = input.size(0) if self.batch_first else input.size(1)

    # layer * direction
    if hx is None:
      num_directions = 2 if self.bidirectional else 1
      hx = var(input.data.new(max_batch_size, self.hidden_size).zero_(), requires_grad=False)
      hx = (hx, hx)
      hx = [[hx for x in range(directions)] for d in range(self.num_layers)]

    # make weights indexable with layer -> direction
    ws = self.all_weights
    if directions == 1:
      ws = [ [w] for w in ws ]
    else:
      ws = [ [ws[l*2], ws[l*2+1]] for l in range(self.num_layers) ]

    # make input batch-first, separate into timeslice wise chunks
    input = input if self.batch_first else input.transpose(0, 1)
    os = [[input[:, i, :] for i in range(timesteps)] for d in range(directions)]
    if directions > 1:
      os[1].reverse()

    for time in range(timesteps):
      for layer in range(self.num_layers):
        for direction in range(directions):

          if self.bias:
            (w_ih, w_hh, b_ih, b_hh) = ws[layer][direction]
          else:
            (w_ih, w_hh) = ws[layer][direction]
            b_ih = None
            b_hh = None

          hy, cy = SubLSTMCellF(os[direction][time], hx[layer][direction], w_ih, w_hh, b_ih, b_hh)
          hx[layer][direction] = (hy, cy)
          os[direction][time] = hy

        if directions > 1:
          os[0][time] = T.cat([ os[d][time] for d in range(directions) ], -1)
          os[1][time] = os[0][time]

    output = T.stack([T.stack(o, 1) for o in os])
    output = T.cat(output, -1) if self.bidirectional else output[0]
    output = output if self.batch_first else output.transpose(0, 1)

    if is_packed:
      output = pack(output, batch_sizes)
    return output, hx

  def __repr__(self):
    s = '{name}({input_size}, {hidden_size}'
    if self.num_layers != 1:
      s += ', num_layers={num_layers}'
    if self.bias is not True:
      s += ', bias={bias}'
    if self.batch_first is not False:
      s += ', batch_first={batch_first}'
    if self.dropout != 0:
      s += ', dropout={dropout}'
    if self.bidirectional is not False:
      s += ', bidirectional={bidirectional}'
    s += ')'
    return s.format(name=self.__class__.__name__, **self.__dict__)

  def __setstate__(self, d):
    super(SubLSTM, self).__setstate__(d)
    self.__dict__.setdefault('_data_ptrs', [])
    if 'all_weights' in d:
      self._all_weights = d['all_weights']
    if isinstance(self._all_weights[0][0], str):
      return
    num_layers = self.num_layers
    num_directions = 2 if self.bidirectional else 1
    self._all_weights = []
    for layer in range(num_layers):
      for direction in range(num_directions):
        suffix = '_reverse' if direction == 1 else ''
        weights = ['weight_ih_l{}{}', 'weight_hh_l{}{}', 'bias_ih_l{}{}', 'bias_hh_l{}{}']
        weights = [x.format(layer, suffix) for x in weights]
        if self.bias:
          self._all_weights += [weights]
        else:
          self._all_weights += [weights[:2]]

  @property
  def all_weights(self):
    return [[getattr(self, weight) for weight in weights] for weights in self._all_weights]



#!/usr/bin/env python3

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pytorch-sublstm',

    version='0.0.2',

    description='Differentiable Neural Computer, for Pytorch',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/ixaxaar/pytorch-sublstm',

    # Author details
    author='Russi Chatterjee',
    author_email='root@ixaxaar.in',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='cortical microcircuit pytorch sublstm',

    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'tasks']),

    install_requires=['torch', 'numpy'],

    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },

    python_requires='>=3',
)



#!/usr/bin/env python3

import pytest
import numpy as np

import torch.nn as nn
import torch as T
from torch.autograd import Variable as var
import torch.nn.functional as F
from torch.nn.utils import clip_grad_norm
import torch.optim as optim
import numpy as np

import sys
import os
import math
import time
sys.path.insert(0, '.')

from subLSTM.functional import SubLSTMCell as SubLSTMCellF
from subLSTM.nn import SubLSTMCell


def test_cell():
  hidden_size = 20
  input_size = 10

  for bias in (True, False):
    input = var(T.randn(3, input_size))
    hx = var(T.randn(3, hidden_size))
    cx = var(T.randn(3, hidden_size))

    cell = SubLSTMCell(input_size, hidden_size, bias=bias)

    for i in range(6):
      (hx, cx) = cell(input, (hx, cx))

    hx.sum().backward()
    assert hx.size() == T.Size([3, hidden_size])
    assert cx.size() == T.Size([3, hidden_size])



#!/usr/bin/env pythonbatch_size

import pytest
import numpy as np

import torch.nn as nn
import torch as T
from torch.autograd import Variable as var
import torch.nn.functional as F
from torch.nn.utils import clip_grad_norm
import torch.optim as optim
import numpy as np

import sys
import os
import math
import time
sys.path.insert(0, '.')

from subLSTM.nn import SubLSTM


def test_rnn():
  hidden_size = 20
  input_size = 10
  seq_len = 5
  batch_size = 7

  for bias in (True, False):
    for batch_first in (True, False):
      input = var(T.randn(batch_size, seq_len, input_size)) if batch_first else var(T.randn(seq_len, batch_size, input_size))
      hx = None
      rnn = SubLSTM(input_size, hidden_size, num_layers=2, bias=bias, batch_first=batch_first)

      outputs = []
      for i in range(6):
        output, hx = rnn(input, hx)
        outputs.append(output)

      T.stack(outputs).sum().backward()

      assert hx[-1][-1][0].size() == T.Size([batch_size, hidden_size])
      assert hx[-1][-1][1].size() == T.Size([batch_size, hidden_size])



def test_rnn_bidirectional():
  hidden_size = 20
  input_size = 10
  seq_len = 5
  batch_size = 7

  for bias in (True, False):
    for batch_first in (True, False):
      input = var(T.randn(batch_size, seq_len, input_size)) if batch_first else var(T.randn(seq_len, batch_size, input_size))
      hx = None
      rnn = SubLSTM(input_size, hidden_size, num_layers=3, bias=bias, batch_first=batch_first, bidirectional=True)

      outputs = []
      for i in range(6):
        output, hx = rnn(input, hx)
        outputs.append(output)

      T.stack(outputs).sum().backward()

      assert hx[-1][-1][0].size() == T.Size([batch_size, hidden_size])
      assert hx[-1][-1][1].size() == T.Size([batch_size, hidden_size])

